/**
 *  PacDasher application. For explanation of this class, see below. 
 *  Copyright (c) 2003 James McCabe. Email: jamesa@oranda.com 
 *  http://www.oranda.com/java/PacDasher/
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
package com.oranda.pacdasher;

import com.oranda.pacdasher.ui.Controller;
import com.oranda.pacdasher.ui.GUI;
import com.oranda.pacdasher.uimodel.UIModel;
import com.oranda.pacdasher.uimodel.util.UIModelConsts;
import com.oranda.util.Output;

import java.awt.Toolkit;

/*
 * Animation class contains main game loop: kicked off by PacDasherMain
 * Thread calls moveAndDraw repeatedly, with a delay, and modifies
 * the State.
 */
public class Animation 
{
    long frameInterval;
    long curTime;
        
    Controller controller;
    
    public Animation(Controller controller, int frameInterval) 
    {    
        this.frameInterval = frameInterval;
        this.controller = controller;
        
        
        AnimationTask animationTask = new AnimationTask();
        Thread t = new Thread(animationTask);        
        t.setPriority(Thread.MAX_PRIORITY - 1);
        t.start();     
        
        // input thread
        Thread.currentThread().setPriority(Thread.MAX_PRIORITY);   
    }

    
    class AnimationTask implements Runnable 
    {
        public void run() 
        {
        	
            PerformanceStats stats = new PerformanceStats();
            stats.initialize();
                                  
            while (true)
            {                
                runSetOfFrames(stats);
                debugPerformanceSummary(stats);
           }
        }
        
        private void runSetOfFrames(PerformanceStats stats)
        {
        	stats.initializeForSet();
            
            while (stats.curFrameNum < PerformanceStats.NUM_FRAMES_PER_SET)
            {                   
                controller.moveAndDraw(); // main functional line in this class
                                                         
                stats.curFrameNum++;
                processState(stats);                                                    
            }	        	
        }
        
        private void processState(PerformanceStats stats)
        {
            /*
                 This flow could be formalized in a Finite State Machine
                 here, but it is not.
                 
                 States: START_STATE        until maze created
                         GAME_READY_STATE    until after DELAY_GAME_READY
                         NORMAL_STATE        until capture of ghost
                         CAPTURE_STATE    until after DELAY_CAPTURE
                         NORMAL_STATE        until after capture of PacDasher
                         READY_STATE        until after DELAY_READY
                         NORMAL_STATE        until isMazeFinished is true
                         BEFORE_LEVEL    until after DELAY_BEFORE_LEVEL
                         READY_STATE        until isGameOver is true
                         GAME_OVER_STATE    until after key pressed
                         GAME_READY_STATE see above
            */
            int state = Controller.getState();
            if (state == State.START_STATE)
            {
                UIModel.getInstance().startNewGame();
                Controller.setState(State.GAME_READY_STATE);
                GUI.getInstance().requestFocus();
            }
            if (state == State.GAME_READY_STATE)
            {
                delay(UIModelConsts.DELAY_GAME_READY, stats, true);
                Controller.setState(State.NORMAL_STATE);
                stats.resetTime();
            }
            else if (state == State.NORMAL_STATE)
            {
                delay(frameInterval, stats, false); 
            }
            else if (state == State.BEFORE_LEVEL_STATE)
            {
                UIModel.getInstance().goToNewLevel();
                delay(UIModelConsts.DELAY_BEFORE_LEVEL, stats, true);                               
                Controller.setState(State.READY_STATE);
            }
            else if (state == State.READY_STATE)
            {
                delay(UIModelConsts.DELAY_READY, stats, true);
                Controller.setState(State.NORMAL_STATE);
                stats.resetTime();
                GUI.getInstance().requestFocus();
            }
            else if (state == State.GAME_OVER_STATE)
            {
            	GUI.getInstance().clearLastKeyEvent();
                delay(UIModelConsts.DELAY_GAME_OVER, stats, true);
            	// wait for key press
            }    
            else if (state == State.CAPTURE_STATE)
            {
                Toolkit.getDefaultToolkit().beep();
                delay(UIModelConsts.DELAY_CAPTURE, stats, true);
                Controller.setState(State.NORMAL_STATE);
                stats.resetTime();
            }                           
                        
        }
    
        protected void debugPerformanceSummary(PerformanceStats stats)
        {
            if (Output.isDebugging())
    		{ 
                long endTime = System.currentTimeMillis();
                if (endTime == stats.beginTime) // avoid div by zero 
                {
                    endTime++;
                }
                long framesPerSec = (PerformanceStats.NUM_FRAMES_PER_SET * 1000)/(endTime - stats.beginTime);
                float millisPerFrame = 1000/(float)framesPerSec;
                long totalMemory = Runtime.getRuntime().totalMemory();
                    
                Output.print(Output.DEBUG, "");
                Output.print(Output.DEBUG, "totalMemory = " + totalMemory 
                    + " freeMemory " + stats.freeMemory);
                Output.print(Output.DEBUG, "bytes used per frame " 
                    + stats.bytesUsedForSet/PerformanceStats.NUM_FRAMES_PER_SET);
    
                Output.print(Output.DEBUG, "framesPerSec = " + framesPerSec);
                Output.print(Output.DEBUG, "millisPerFrame = " + millisPerFrame);
                Output.print(Output.DEBUG, "millis minus delay = " 
                            + (millisPerFrame - frameInterval));
    
                Output.print(Output.DEBUG, "numTooLongs = " + stats.numTooLongs + 
                    " gcCount " + stats.gcCount);
    		}
    	}
            
    
        protected void delay(long milliseconds, PerformanceStats stats, 
            boolean force)
        {
              
            stats.idealTime += milliseconds;
                    	
            long curTime = System.currentTimeMillis();
            long correctDelay = 0;
                    
            if (force == true)
            {
            	correctDelay = milliseconds;
            }
            else
            {
            	// frame rate control!
            	correctDelay = stats.idealTime - curTime;
            }
    
            /*if (correctDelay < 0)
            {
                Output.print(Output.DEBUG, "idealTime " + stats.idealTime);
                Output.print(Output.DEBUG, "curTime " + curTime); 
                Output.print(Output.DEBUG, "correctDelay " + correctDelay);
            }*/	
            if (correctDelay > 0)
            {           	        	 	    
                if (correctDelay >= 500)
                {
                    System.gc();
                }
                try
                {
                	//Output.print(Output.DEBUG, "sleeping " + correctDelay);                               
                    Thread.currentThread().sleep(correctDelay);
                }
                catch (InterruptedException e)
                {
                    Output.print(Output.ERROR, "Interrupted while sleeping " + milliseconds);
                } 
            }
            else
            {
            	 // Give input thread a chance anytway
            	 Thread.currentThread().yield();
            }       
            debugPerformance(stats, (-correctDelay));
            if (correctDelay <= -60)
            {
                // unrecoverable overrun, e.g. from a full paint
            	stats.resetTime();
            	correctDelay = milliseconds;
            }    
        }

    
        protected void debugPerformance(PerformanceStats stats, long overrun)
        {        	
            if (Output.isDebugging())
    	    {
    	    	//long curTime = System.currentTimeMillis();
    	    	//long delay = curTime - stats.prevTime;
    	    	  
                if (overrun > frameInterval)
                {
                    Output.print(Output.DEBUG, "****************** "
                        + "too long! frame " 
                        + stats.curFrameNum + " overrun " + overrun
                        + " idealTime " + stats.idealTime);
                    stats.numTooLongs++; 
                }
                stats.freeMemory = Runtime.getRuntime().freeMemory();
                
                long memoryUsed = stats.lastFreeMemory - stats.freeMemory;
                if (memoryUsed < 0)
                {
                	Output.print(Output.DEBUG, "memoryFreed, frame " 
                	    + stats.curFrameNum + " freed " + (-memoryUsed));
                    stats.gcCount++;
                }
                else 
                {
                	stats.bytesUsedForSet = stats.bytesUsedForSet + memoryUsed;
                }
                
                stats.lastFreeMemory = stats.freeMemory;
                //stats.prevTime = curTime;
            }
        }
    }    


    class PerformanceStats
    {
    	static final int NUM_FRAMES_PER_SET = 500;
        long beginTime; 
        //long prevTime;
        long tooLong; 
        int numTooLongs;
        long bytesUsedForSet;
        long idealTime;
        long freeMemory;
        long lastFreeMemory;
        long gcCount;
        int curFrameNum;
        
        public void resetTime()
        {
        	idealTime = System.currentTimeMillis();
        	Output.print(Output.DEBUG, "idealTime was reset to " + idealTime);
        } 
        
        public void initialize()
        {
            freeMemory = 0;
            lastFreeMemory = freeMemory;
            gcCount = 0;            
            tooLong = frameInterval * 2;
            //prevTime = beginTime;
            idealTime = beginTime; 
        }
        
        public void initializeForSet()
        {
        	beginTime = System.currentTimeMillis();
        	curFrameNum = 0;
        	numTooLongs = 0;
            bytesUsedForSet = 0;        	
        }

    }
    				                        
    /*
     * for testing
     */
    public static void main(String args[]) 
    {

    }

}
