/**
 *  PacDasher application. For explanation of this class, see below. 
 *  Copyright (c) 2003 James McCabe. Email: jamesa@oranda.com 
 *  http://www.oranda.com/java/PacDasher/
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
package com.oranda.pacdasher.uimodel;

import com.oranda.pacdasher.uimodel.util.*;
import com.oranda.util.Output;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class GhostStrategy
{
    protected static ArrayList possibleDirections;
    protected Ghost ghost;
    protected Random randomGenerator;
    protected int virtualTimeEnteredFlight;
    
    public final static int NORMAL_STATE = 0;
    public final static int SCATTER_STATE = 1;
    public final static int FLIGHT_STATE = 2;
    
    protected int state;
    
    static
    {
        possibleDirections = new ArrayList(4);
        possibleDirections.add(DirectionCode.UP);
        possibleDirections.add(DirectionCode.LEFT);
        possibleDirections.add(DirectionCode.DOWN);
        possibleDirections.add(DirectionCode.RIGHT);
    }
    
    public GhostStrategy(Ghost ghost)
    {
        this.ghost = ghost;        
        randomGenerator = new Random();
    }


    public void move(PosAndDirection posAndDirection,
        XY pacDasherXy, DirectionCode pacDasherDirectionCode)
    {
        MoveAttempt moveAttempt;
        if (state == FLIGHT_STATE)
         {
             // if flight time is up, restore to normal
             if (getVirtualTimeSinceFlight() > Ghost.TIME_FLIGHT_TOTAL)
             {
                 setState(NORMAL_STATE);
             }
             // halve speed in flight state
             else if (UIModel.getVirtualTime()%2 == 1)
             {
                 return; // i.e. don't move
             }
         } 
         // every once in a while enter scatter state
         else if (getState() != SCATTER_STATE &&
             UIModel.getVirtualTime()%Ghost.PERIOD_SCATTER <= Ghost.TIME_SCATTER)
        {
            setState(SCATTER_STATE);
        }
        else if (getState() == SCATTER_STATE &&
            UIModel.getVirtualTime()%Ghost.PERIOD_SCATTER > Ghost.TIME_SCATTER)
        {
            setState(NORMAL_STATE);
           }
   
        if (state == NORMAL_STATE)
        {
            moveAttempt = NormalMoveAttempt.getInstance();
        }
        else if (state == SCATTER_STATE)
        {
            moveAttempt = ScatterMoveAttempt.getInstance();
        }
        else //if (state == FLIGHT_STATE)
        {
            moveAttempt = FlightMoveAttempt.getInstance();
        }
        
        tryMove(posAndDirection, pacDasherXy, 
            pacDasherDirectionCode, moveAttempt);

        //Output.print(Output.DEBUG, "GhostStrategy: after move " + posAndDirection.getXY());
    }

    public void tryMove(PosAndDirection posAndDirection,
        XY pacDasherXy, DirectionCode pacDasherDirectionCode,
        MoveAttempt moveAttempt)
    {
        XY xy = (XY) ghost.getXY().clone();

        if ((xy.getX()%UIModelConsts.X_TILE_SIZE == 0 && ghost.isMovingHorizontally())
            || (xy.getY()%UIModelConsts.Y_TILE_SIZE == 0 && ghost.isMovingVertically()))
        {
            //Output.print(Output.DEBUG, "move() xy " + xy);
            //resetPossibleDirections();
            // if at a possible turning point, decide upon a new direction
            tryMoveDirections(posAndDirection,
                (java.util.List) possibleDirections.clone(),
               pacDasherXy, moveAttempt);
            XY newXy = posAndDirection.move();    
        } 
        else 
        {
            XY newXy = posAndDirection.move();
            if (newXy.equals(xy))
            {
                // move unsuccessful  
                // next line includes posAndDirection.move()              
                tryMoveDirections(posAndDirection,
                    (java.util.List) possibleDirections.clone(),
                    pacDasherXy, moveAttempt);    
            }
        }        
    }


    /* 
     * Calls itself recursively until there is a move
     */
    public void tryMoveDirections(PosAndDirection posAndDirection,
        java.util.List allowedDirections, 
        XY pacDasherXy, MoveAttempt moveAttempt)
    {
        //Output.print(Output.DEBUG, "tryMoveDirections " + allowedDirections);
        if (allowedDirections.isEmpty())
        {
            return;
        }
        
        XY xy = (XY) ghost.getXY().clone();
        DirectionCode desiredDirectionCode 
            = moveAttempt.propose(ghost, pacDasherXy, allowedDirections);
                    
        posAndDirection.setDesiredDirectionCode(desiredDirectionCode);
        XY newXy = posAndDirection.move();     
        //Output.print(Output.DEBUG, desiredDirectionCode + " " + xy 
        //    + " " + newXy); 
        if (xy.equals(newXy))
        {
            // move unsuccessful
            allowedDirections.remove(desiredDirectionCode);
            tryMoveDirections(posAndDirection, allowedDirections, 
                pacDasherXy, moveAttempt);
        }
    }

    public int getState()
    {
        return state;
    }
    
    public int getVirtualTimeEnteredFlight()
    {
        return virtualTimeEnteredFlight;
    }
    
    public int getVirtualTimeSinceFlight()
    {
        return UIModel.getVirtualTime() 
            - getVirtualTimeEnteredFlight();
    }

    public void setState(int state)
    {
        //Output.print(Output.DEBUG, "Ghost going to state " + state);
        this.state = state;
        if (state == FLIGHT_STATE)
        {
            markVirtualTimeEnteredFlight();
        }
    }
    
    public void markVirtualTimeEnteredFlight()
    {
        virtualTimeEnteredFlight = UIModel.getVirtualTime();
    }    
    
}
 